from enum import Enum
import typing
import re


class ComparisonComparators(Enum):
    """ Used inside a Comparison Expression to describe a relationship between a field and value. """
    (
        Equal,
        NotEqual,
        GreaterThan,
        LessThan,
        GreaterThanOrEqual,
        LessThanOrEqual,
        In,
        Like,
        Matches,
        IsSuperSet,
        IsSubSet
    ) = range(11)

    def __repr__(self):
        return self._name_


class ComparisonExpressionOperators(Enum):
    """ Used to combine two Comparison Expressions together inside an ObservationExpression"""
    (
        And,
        Or
    ) = range(2)

    def __repr__(self):
        return self._name_


class ObservationOperators(Enum):
    """ Used to combine two or more ObservationExpressions."""
    (
        And,
        Or,
        FollowedBy
    ) = range(3)

    def __repr__(self):
        return self._name_


class STIX2Value:
    pass


class SetValue(STIX2Value):
    def __init__(self):
        self.values = []
        self.is_open = True

    def append(self, value):
        if self.is_open:
            self.values.append(value)
        else:
            raise RuntimeError("Cannot append to closed Set")

    def close(self):
        self.is_open = False

    def element_iterator(self):
        for value in self.values:
            yield str(value)

    def __str__(self):
        return "({})".format(str(self.values).lstrip('[').rstrip(']'))


class BaseComparisonExpression:
    pass


class ComparisonExpression(BaseComparisonExpression):
    def __init__(self, object_path, value, comparator: ComparisonComparators, negated: bool=False):
        if not isinstance(comparator, ComparisonComparators):
            raise RuntimeWarning("{} is not a ComparisonComparator".format(comparator))
        self.object_path = object_path
        self.value = value
        self.comparator = comparator
        self.negated = negated

    def __repr__(self):
        return "ComparisonExpression({field} {comparator} {value})".format(comparator=self.comparator,
                                                                           field=self.object_path,
                                                                           value=self.value)


class CombinedComparisonExpression(BaseComparisonExpression):
    def __init__(self, expr1: BaseComparisonExpression, expr2: BaseComparisonExpression,
                 operator: ComparisonExpressionOperators) -> None:
        if not all((isinstance(expr1, BaseComparisonExpression), isinstance(expr2, BaseComparisonExpression),
                    isinstance(operator, ComparisonExpressionOperators))):
            raise RuntimeWarning("{} constructor called with wrong types".format(__class__)) # noqa: F821
        self.expr1 = expr1
        self.expr2 = expr2
        self.operator = operator

    def __repr__(self) -> str:
        return "CombinedComparisonExpression({expr1} {operator} {expr2})".format(expr1=self.expr1,
                                                                                 operator=self.operator,
                                                                                 expr2=self.expr2)


class BaseObservationExpression:
    pass


class ObservationExpression(BaseObservationExpression):
    def __init__(self, comparison_expression: BaseComparisonExpression) -> None:
        if not isinstance(comparison_expression, BaseComparisonExpression):
            raise RuntimeWarning("{} constructor called with wrong types".format(__class__)) # noqa: F821
        self.comparison_expression = comparison_expression

    def __repr__(self) -> str:
        return "ObservationExpression({expr})".format(expr=self.comparison_expression)


class CombinedObservationExpression(BaseObservationExpression):
    # This method is recursively hit when there are more than two base observation expressions
    # A CombinedObservationExpression will only contain up to two ObservationExpressions, joined by an ObservationOperator
    def __init__(self, expr1: BaseObservationExpression, expr2: BaseObservationExpression,
                 operator: ObservationOperators) -> None:
        self.expr1 = expr1
        self.expr2 = expr2
        self.operator = operator

        self.__check_instances()

    def __check_instances(self):

        if hasattr(self.expr1, 'observation_expression'):
            wrong_type = not isinstance(self.expr1.observation_expression, BaseObservationExpression)
        else:
            wrong_type = not isinstance(self.expr1, BaseObservationExpression)

        if wrong_type:
            raise RuntimeWarning("{} constructor called with wrong types".format(__class__)) # noqa: F821

        if hasattr(self.expr2, 'observation_expression'):
            wrong_type = not isinstance(self.expr2.observation_expression, BaseObservationExpression)
        else:
            wrong_type = not isinstance(self.expr2, BaseObservationExpression)
        if wrong_type:
            raise RuntimeWarning("{} constructor called with wrong types".format(__class__)) # noqa: F821

        if not isinstance(self.operator, ObservationOperators):
            raise RuntimeWarning("{} constructor called with wrong types".format(__class__)) # noqa: F821

    def __repr__(self) -> str:
        return "CombinedObservationExpression({expr1} {operator} {expr2})".format(expr1=self.expr1,
                                                                                  operator=self.operator,
                                                                                  expr2=self.expr2)


class BaseQualifier:
    pass


class Qualifier(BaseQualifier):
    def __init__(self, qualifier, observation_expression: BaseObservationExpression) -> None:
        if not isinstance(observation_expression, BaseObservationExpression):
            raise RuntimeWarning("{} constructor called with wrong types".format(__class__)) # noqa: F821
        self.qualifier = qualifier
        self.observation_expression = observation_expression

    def __repr__(self) -> str:
        return "{observation_expression} Qualifier({qualifier})".format(observation_expression=self.observation_expression, qualifier=self.qualifier)

class StartStopQualifier(Qualifier):
    def __init__(self, qualifier, observation_expression: BaseObservationExpression, start: str, stop: str) -> None:
        if not isinstance(observation_expression, BaseObservationExpression):
            raise RuntimeWarning("{} constructor called with wrong types".format(__class__)) # noqa: F821
        self.qualifier = qualifier
        self.observation_expression = observation_expression
        self.start = start
        self.stop = stop

        pattern = "^t'\d{4}(-\d{2}){2}T\d{2}(:\d{2}){2}(\.\d+)?Z'$"
        match = re.search(pattern, start)
        if not bool(match):
            raise RuntimeError("Invalid STIX timestamp {}".format(start))

        match = re.search(pattern, stop)
        if not bool(match):
            raise RuntimeError("Invalid STIX timestamp {}".format(stop))

    def __repr__(self) -> str:
        return "{observation_expression} StartStopQualifier({qualifier}, start={start}, stop={stop})".format(observation_expression=self.observation_expression, qualifier=self.qualifier, start=self.start, stop=self.stop)

class Pattern:
    def __init__(self, expression: BaseObservationExpression, qualifier=None) -> None:
        self.expression = expression

    def __repr__(self) -> str:
        return "Pattern[{expression}]".format(expression=self.expression)



# Generated from STIXPattern.g4 by ANTLR 4.7.1
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .STIXPatternParser import STIXPatternParser
else:
    from STIXPatternParser import STIXPatternParser

# This class defines a complete generic visitor for a parse tree produced by STIXPatternParser.

class STIXPatternVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by STIXPatternParser#pattern.
    def visitPattern(self, ctx:STIXPatternParser.PatternContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by STIXPatternParser#observationExpressions.
    def visitObservationExpressions(self, ctx:STIXPatternParser.ObservationExpressionsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by STIXPatternParser#observationExpressionOr.
    def visitObservationExpressionOr(self, ctx:STIXPatternParser.ObservationExpressionOrContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by STIXPatternParser#observationExpressionAnd.
    def visitObservationExpressionAnd(self, ctx:STIXPatternParser.ObservationExpressionAndContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by STIXPatternParser#observationExpressionRepeated.
    def visitObservationExpressionRepeated(self, ctx:STIXPatternParser.ObservationExpressionRepeatedContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by STIXPatternParser#observationExpressionSimple.
    def visitObservationExpressionSimple(self, ctx:STIXPatternParser.ObservationExpressionSimpleContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by STIXPatternParser#observationExpressionCompound.
    def visitObservationExpressionCompound(self, ctx:STIXPatternParser.ObservationExpressionCompoundContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by STIXPatternParser#observationExpressionWithin.
    def visitObservationExpressionWithin(self, ctx:STIXPatternParser.ObservationExpressionWithinContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by STIXPatternParser#observationExpressionStartStop.
    def visitObservationExpressionStartStop(self, ctx:STIXPatternParser.ObservationExpressionStartStopContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by STIXPatternParser#comparisonExpressionAnd_.
    def visitComparisonExpressionAnd_(self, ctx:STIXPatternParser.ComparisonExpressionAnd_Context):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by STIXPatternParser#comparisonExpressionOred.
    def visitComparisonExpressionOred(self, ctx:STIXPatternParser.ComparisonExpressionOredContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by STIXPatternParser#comparisonExpressionAndPropTest.
    def visitComparisonExpressionAndPropTest(self, ctx:STIXPatternParser.ComparisonExpressionAndPropTestContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by STIXPatternParser#comparisonExpressionAnded.
    def visitComparisonExpressionAnded(self, ctx:STIXPatternParser.ComparisonExpressionAndedContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by STIXPatternParser#propTestEqual.
    def visitPropTestEqual(self, ctx:STIXPatternParser.PropTestEqualContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by STIXPatternParser#propTestOrder.
    def visitPropTestOrder(self, ctx:STIXPatternParser.PropTestOrderContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by STIXPatternParser#propTestSet.
    def visitPropTestSet(self, ctx:STIXPatternParser.PropTestSetContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by STIXPatternParser#propTestLike.
    def visitPropTestLike(self, ctx:STIXPatternParser.PropTestLikeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by STIXPatternParser#propTestRegex.
    def visitPropTestRegex(self, ctx:STIXPatternParser.PropTestRegexContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by STIXPatternParser#propTestIsSubset.
    def visitPropTestIsSubset(self, ctx:STIXPatternParser.PropTestIsSubsetContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by STIXPatternParser#propTestIsSuperset.
    def visitPropTestIsSuperset(self, ctx:STIXPatternParser.PropTestIsSupersetContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by STIXPatternParser#propTestParen.
    def visitPropTestParen(self, ctx:STIXPatternParser.PropTestParenContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by STIXPatternParser#orderingComparator.
    def visitOrderingComparator(self, ctx:STIXPatternParser.OrderingComparatorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by STIXPatternParser#stringLiteral.
    def visitStringLiteral(self, ctx:STIXPatternParser.StringLiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by STIXPatternParser#startStopQualifier.
    def visitStartStopQualifier(self, ctx:STIXPatternParser.StartStopQualifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by STIXPatternParser#withinQualifier.
    def visitWithinQualifier(self, ctx:STIXPatternParser.WithinQualifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by STIXPatternParser#repeatedQualifier.
    def visitRepeatedQualifier(self, ctx:STIXPatternParser.RepeatedQualifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by STIXPatternParser#objectPath.
    def visitObjectPath(self, ctx:STIXPatternParser.ObjectPathContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by STIXPatternParser#objectType.
    def visitObjectType(self, ctx:STIXPatternParser.ObjectTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by STIXPatternParser#firstPathComponent.
    def visitFirstPathComponent(self, ctx:STIXPatternParser.FirstPathComponentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by STIXPatternParser#indexPathStep.
    def visitIndexPathStep(self, ctx:STIXPatternParser.IndexPathStepContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by STIXPatternParser#pathStep.
    def visitPathStep(self, ctx:STIXPatternParser.PathStepContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by STIXPatternParser#keyPathStep.
    def visitKeyPathStep(self, ctx:STIXPatternParser.KeyPathStepContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by STIXPatternParser#setLiteral.
    def visitSetLiteral(self, ctx:STIXPatternParser.SetLiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by STIXPatternParser#primitiveLiteral.
    def visitPrimitiveLiteral(self, ctx:STIXPatternParser.PrimitiveLiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by STIXPatternParser#orderableLiteral.
    def visitOrderableLiteral(self, ctx:STIXPatternParser.OrderableLiteralContext):
        return self.visitChildren(ctx)



del STIXPatternParser


from stix_shifter_utils.stix_translation.src.patterns.pattern_objects import ObservationExpression, ComparisonExpression, \
    ComparisonExpressionOperators, ComparisonComparators, Pattern, \
    CombinedComparisonExpression, CombinedObservationExpression, ObservationOperators
import datetime
from stix_shifter_utils.stix_translation.src.utils.transformers import DateTimeToUnixTimestamp
import re


class PatternTranslator:
    comparator_lookup = {
        ComparisonExpressionOperators.And: "AND",
        ComparisonExpressionOperators.Or: "OR",
        ComparisonComparators.GreaterThan: ">",
        ComparisonComparators.GreaterThanOrEqual: ">=",
        ComparisonComparators.LessThan: "<",
        ComparisonComparators.LessThanOrEqual: "<=",
        ComparisonComparators.Equal: "=",
        ComparisonComparators.NotEqual: "!=",
        ComparisonComparators.Like: "LIKE",
        ComparisonComparators.In: "IN",
        ComparisonComparators.Matches: 'MATCHES',
        ObservationOperators.Or: 'OR',
        # Treat AND's as OR's -- Unsure how two ObsExps wouldn't cancel each other out.
        ObservationOperators.And: 'OR',
        ObservationOperators.FollowedBy: 'FOLLOWEDBY',
        ComparisonComparators.IsSuperSet: 'ISSUPERSET',
        ComparisonComparators.IsSubSet: 'ISSUBSET'
    }

    def __init__(self, pattern: Pattern, time_range):
        self.parsed_pattern = []
        # Set times based on default time_range or what is in the options
        # START STOP qualifiers will override this
        end_time = datetime.datetime.utcnow()
        self.end_time = DateTimeToUnixTimestamp.transform(end_time)
        go_back_in_minutes = datetime.timedelta(minutes=time_range)
        start_time = end_time - go_back_in_minutes
        self.start_time = DateTimeToUnixTimestamp.transform(start_time)
        self.qualifier_timerange_override = False
        self.parse_expression(pattern)

    def _parse_expression(self, expression, qualifier=None) -> str:
        if isinstance(expression, ComparisonExpression):  # Base Case
            # Resolve STIX Object Path to a field in the target Data Model
            stix_object, stix_field = expression.object_path.split(':')
            comparator = self.comparator_lookup[expression.comparator]
            if expression.negated:
                comparator = 'NOT ' + comparator
            if qualifier is not None:
                self._convert_qualifier_times_to_unix_times(qualifier)
            self.parsed_pattern.append({'attribute': expression.object_path, 'comparison_operator': comparator, 'value': expression.value})
        elif isinstance(expression, CombinedComparisonExpression):
            if qualifier is not None:
                self._convert_qualifier_times_to_unix_times(qualifier)
            self._parse_expression(expression.expr1)
            self._parse_expression(expression.expr2)

        elif isinstance(expression, ObservationExpression):
            self._parse_expression(expression.comparison_expression, qualifier)

        elif hasattr(expression, 'qualifier') and hasattr(expression, 'observation_expression'):
            if isinstance(expression.observation_expression, CombinedObservationExpression):
                self._parse_expression(expression.observation_expression.expr1)
                self._parse_expression(expression.observation_expression.expr2, expression.qualifier)
            else:
                self._parse_expression(expression.observation_expression.comparison_expression, expression.qualifier)
        elif isinstance(expression, CombinedObservationExpression):
            self._parse_expression(expression.expr1)
            self._parse_expression(expression.expr2)
        elif isinstance(expression, Pattern):
            self._parse_expression(expression.expression)
        else:
            raise RuntimeError("Unknown Recursion Case for expression={}, type(expression)={}".format(
                expression, type(expression)))

    def _convert_qualifier_times_to_unix_times(self, qualifier):
        split_object = qualifier.split("'")
        start_time = split_object[1]
        end_time = split_object[3]
        # Add subseconds to timestamp if it's missing
        pattern = "\.\d+Z$"
        if not bool(re.search(pattern, start_time)):
            start_time = re.sub('Z$', '.000Z', start_time)
        if not bool(re.search(pattern, end_time)):
            end_time = re.sub('Z$', '.000Z', end_time)
        # convert from string format '2019-01-18T08:30:16.227Z'
        start_time = datetime.datetime.strptime(start_time, '%Y-%m-%dT%H:%M:%S.%fZ')
        end_time = datetime.datetime.strptime(end_time, '%Y-%m-%dT%H:%M:%S.%fZ')
        start_time = DateTimeToUnixTimestamp.transform(start_time)
        end_time = DateTimeToUnixTimestamp.transform(end_time)
        if not self.qualifier_timerange_override or self.start_time > start_time:
            self.start_time = start_time
        if not self.qualifier_timerange_override or self.end_time < end_time:
            self.end_time = end_time
        self.qualifier_timerange_override = True

    def parse_expression(self, pattern: Pattern):
        return self._parse_expression(pattern)


def parse_stix(pattern: Pattern, time_range):
    x = PatternTranslator(pattern, time_range)
    return {'parsed_stix': x.parsed_pattern, 'start_time': x.start_time, 'end_time': x.end_time}



from .base_ping_connector import BasePingConnector
from .base_query_connector import BaseQueryConnector
from .base_status_connector import BaseStatusConnector
from .base_delete_connector import BaseDeleteConnector
from .base_results_connector import BaseResultsConnector
from .base_connector import BaseConnector


class BaseSyncConnector(BaseConnector):
    def create_query_connection(self, query):
        return {"success": True, "search_id": query}

    def create_status_connection(self, search_id):
        return {"success": True, "status": "COMPLETED", "progress": 100}




from stix_shifter.stix_translation import stix_translation
from stix_shifter.stix_transmission import stix_transmission
from flask import request
import json


class ProxyHost():

    def __init__(self):
        self.request_args = request.get_json(force=True)
        self.connection = self.request_args.get("connection")
        self.configuration = self.request_args.get("configuration")
        if self.connection:
            self.options = self.connection.get("options", {})
        else:
            self.connection = self.request_args.get("options", {})

    def transform_query(self):
        query = self.request_args["query"]
        translation_module = self.connection['type'].lower()
        translation = stix_translation.StixTranslation()
        dsl = translation.translate(translation_module, 'query', '{}', query, self.connection)
        return json.dumps(dsl['queries'])

    def translate_results(self, data_source_identity_object):
        data_source_results = json.dumps(self.request_args["results"] )

        print(data_source_results)
        translation_module = self.connection['type'].lower()
        translation = stix_translation.StixTranslation()
        dsl = translation.translate(translation_module, 'results', data_source_identity_object, data_source_results, self.connection)
        return json.dumps(dsl)

    def create_query_connection(self):
        query = self.request_args["query"]
        transmission_module = self.connection['type'].lower()
        transmission = stix_transmission.StixTransmission(transmission_module, self.connection, self.configuration)
        return json.dumps(transmission.query(query))

    def create_status_connection(self):
        search_id = self.request_args["search_id"]
        transmission_module = self.connection['type'].lower()
        transmission = stix_transmission.StixTransmission(transmission_module, self.connection, self.configuration)
        return json.dumps(transmission.status(search_id))

    def create_results_connection(self):
        search_id = self.request_args["search_id"]
        offset = self.request_args["offset"]
        length = self.request_args["length"]
        transmission_module = self.connection['type'].lower()
        transmission = stix_transmission.StixTransmission(transmission_module, self.connection, self.configuration)
        return json.dumps(transmission.results(search_id, offset, length))

    def delete_query_connection(self):
        search_id = self.request_args["search_id"]
        transmission_module = self.connection['type'].lower()
        transmission = stix_transmission.StixTransmission(transmission_module, self.connection, self.configuration)
        return json.dumps(transmission.delete(search_id))

    def ping_connection(self):
        transmission_module = self.connection['type'].lower()
        transmission = stix_transmission.StixTransmission(transmission_module, self.connection, self.configuration)
        return json.dumps(transmission.ping())

    def is_async(self):
        transmission_module = self.connection['type'].lower()
        transmission = stix_transmission.StixTransmission(transmission_module, self.connection, self.configuration)
        return "{}".format(transmission.is_async())



from abc import ABCMeta, abstractmethod


class BaseQueryTranslator(object, metaclass=ABCMeta):

    def __init__(self, dialect=None, default_mapping_file_path=None):
        self.dialect = dialect
        self.default_mapping_file_path = default_mapping_file_path

    @abstractmethod
    def transform_query(self, data, antlr_parsing_object, data_model_mapper, options, mapping=None):
        """
        Transforms STIX pattern into a different query format. Based on a mapping file
        :param data: STIX pattern to transform into native data source query
        :type data: str
        :param antlr_parsing_object: Antlr parsing objects for the STIX pattern
        :type antlr_parsing_object: object
        :data_model_mapper: Mapping object for the data source
        :type data_model_mapper: object
        :param mapping: The mapping file path to use as instructions on how to transform the given STIX query into another format. This should default to something if one isn't passed in
        :type mapping: str (filepath)
        :return: transformed query string
        :rtype: str
        """
        # if translating STIX pattern to a datasource query...
        pass



import requests
from requests_toolbelt.adapters import host_header_ssl
import sys
import collections
import urllib.parse
import os
import errno

# This is a simple HTTP client that can be used to access the REST API
CLIENT_CERT_NAME = "client_cert.pem"
SERVER_CERT_NAME = "server_cert.pem"


class RestApiClient:
    #cert_verify can be True -- do proper signed cert check, False -- skip all cert checks, or a Cert -- use the proper cleint side cert
    #mutual_auth is in the case the gateway is being used
    def __init__(self, host, port=None, cert=None, headers={}, url_modifier_function=None, cert_verify=True,
                 mutual_auth=False, sni=None):
        server_ip = host
        if port is not None:
            server_ip += ":" + str(port)
        self.server_ip = server_ip
        #sni is none unless we are using a server cert
        self.sni = None
        #Gateway Case -- use client cert cert_verify is None
        if mutual_auth:
            self.server_cert_content = None
            self.server_cert_file_content_exists = False
            self.client_cert_content = CLIENT_CERT_NAME
            self.client_cert_file_content_exists = True
            self.client_cert_file_content = cert
        #verify is true or false
        elif isinstance(cert_verify, bool):
            if cert_verify:
                self.server_cert_content = True
                self.server_cert_file_content_exists = False
                self.client_cert_content = None
                self.client_cert_file_content_exists = False
            else:
                self.server_cert_content = False
                self.server_cert_file_content_exists = False
                self.client_cert_content = None
                self.client_cert_file_content_exists = False
        #server cert provided
        elif isinstance(cert_verify, str):
            self.server_cert_content = SERVER_CERT_NAME
            self.server_cert_file_content_exists = True
            self.server_cert_file_content = cert_verify
            self.client_cert_content = None
            self.client_cert_file_content_exists = False
            if sni is not None:
                self.sni = sni

        self.headers = headers
        self.url_modifier_function = url_modifier_function

    # This method is used to set up an HTTP request and send it to the server
    def call_api(self, endpoint, method, headers=None, params=[], data=None, urldata=None, timeout=None):
        try:

            # convert client cert to file
            if self.client_cert_file_content_exists is True:
                with open(CLIENT_CERT_NAME, 'w') as f:
                    try:
                        f.write(self.client_cert_file_content)
                    except IOError:
                        print('Failed to setup certificate')

            # covnert server cert to file
            if self.server_cert_file_content_exists is True:
                with open(SERVER_CERT_NAME, 'w') as f:
                    try:
                        f.write(self.server_cert_file_content)
                    except IOError:
                        print('Failed to setup certificate')

            url = None
            actual_headers = self.headers.copy()
            if headers is not None:
                for header_key in headers:
                    actual_headers[header_key] = headers[header_key]

            if urldata:
                urldata = urllib.parse.urlencode(urldata)
                if '?' in endpoint:
                    endpoint += '&'
                else:
                    endpoint += '?'
                endpoint += urldata

            if self.url_modifier_function is not None:
                url = self.url_modifier_function(
                    self.server_ip, endpoint, actual_headers)
            else:
                url = 'https://' + self.server_ip + '/' + endpoint
            try:
                call = getattr(requests, method.lower())

                # only use the tool belt session in case of SNI for safety
                if self.sni is not None:
                    session = requests.Session()
                    call = getattr(session, method.lower())
                    session.mount('https://', host_header_ssl.HostHeaderSSLAdapter())
                    actual_headers["Host"] = self.sni

                response = call(url, headers=actual_headers,
                                cert=self.client_cert_content, data=data, verify=self.server_cert_content, timeout=timeout)

                if 'headers' in dir(response) and isinstance(response.headers, collections.Mapping) and 'Content-Type' in response.headers \
                        and "Deprecated" in response.headers['Content-Type']:
                    print("WARNING: " +
                          response.headers['Content-Type'], file=sys.stderr)
                return ResponseWrapper(response)
            except Exception as e:
                print('exception occured during requesting url: ' + str(e))
                raise e
        finally:
            if self.server_cert_file_content_exists is True:
                try:
                    os.remove(SERVER_CERT_NAME)
                except OSError as e:
                    if e.errno != errno.ENOENT:
                        raise
            if self.client_cert_file_content_exists is True:
                try:
                    os.remove(CLIENT_CERT_NAME)
                except OSError as e:
                    if e.errno != errno.ENOENT:
                        raise

    # Simple getters that can be used to inspect the state of this client.
    def get_headers(self):
        return self.headers.copy()

    def get_server_ip(self):
        return self.server_ip


class ResponseWrapper:
    def __init__(self, response):
        self.response = response

    def read(self):
        return self.response.content

    @property
    def bytes(self):
        return self.response.content

    @property
    def code(self):
        return self.response.status_code



from abc import ABCMeta, abstractmethod
from os import path
import json
from stix_shifter_utils.stix_translation.src.utils.exceptions import DataMappingException
import glob


class BaseDataMapper(object, metaclass=ABCMeta):

    def __init__(self, dialect):
        self.dialect = dialect

    def fetch_mapping(self, basepath):
        """
        Fetches STIX-to-datasource mapping JSON from the module's from_stix_map.json file
        :param basepath: path of data source translation module
        :type basepath: str
        """
        try:
            if hasattr(self, 'dialect') and not(self.dialect == None) and not(self.dialect == 'default'):
                filepath = self._fetch_from_stix_mapping_file(basepath)
            else:
                filepath = path.abspath(path.join(basepath, "json", 'from_stix_map.json'))
            map_file = open(filepath).read()
            map_data = json.loads(map_file)
            return map_data
        except Exception as ex:
            print('exception in main():', ex)
            return {}

    def map_field(self, stix_object_name, stix_property_name):
        """
        Maps the STIX object:property pair to any matching data source fields. 
        Mapping is based on a JSON object defined in the data source DataMapper class
        :param stix_object_name: STIX object (ie. url)
        :type stix_object_name: str
        :param stix_property_name: STIX property associated to the object (ie. value)
        :type stix_property_name: str
        :return: A list of 0 or more data source fields that map to a combination of stix_object_name and stix_property_name
        :rtype: list
        """
        if stix_object_name in self.map_data and stix_property_name in self.map_data[stix_object_name]["fields"]:
            return self.map_data[stix_object_name]["fields"][stix_property_name]
        else:
            return []

    def _fetch_from_stix_mapping_file(self, basepath):
        mapping_paths = glob.glob(path.abspath(path.join(basepath, "json", "{}_from_stix*.json".format(self.dialect))))
        return mapping_paths[0]



from .STIXPatternLexer import STIXPatternLexer
from .STIXPatternListener import STIXPatternListener
from .STIXPatternParser import STIXPatternParser
from .STIXPatternVisitor import STIXPatternVisitor



class SearchFeatureNotSupportedError(NotImplementedError):
    """ Use this error when the current features allowed by the search platform make this feature difficult or
    impossible to implement. For instance, analytics that require subsearches or joins are usually going to be
    un-implementable against NoSQL search platforms like ElasticSearch. In the future, we may develop a client-side
    engine to cover these cases."""



import re
import uuid

from . import observable
from stix2validator import validate_instance, print_results
from datetime import datetime


# convert JSON data to STIX object using map_data and transformers

def convert_to_stix(data_source, map_data, data, transformers, options, callback=None):
    bundle = {
        "type": "bundle",
        "id": "bundle--" + str(uuid.uuid4()),
        "spec_version": "2.0",
        "objects": []
    }

    identity_id = data_source['id']
    bundle['objects'] += [data_source]

    ds2stix = DataSourceObjToStixObj(identity_id, map_data, transformers, options, callback)

    # map data list to list of transformed objects
    results = list(map(ds2stix.transform, data))

    bundle["objects"] += results

    return bundle


class DataSourceObjToStixObj:

    def __init__(self, identity_id, ds_to_stix_map, transformers, options, callback=None):
        self.identity_id = identity_id
        self.ds_to_stix_map = ds_to_stix_map
        self.transformers = transformers
        self.options = options
        self.callback = callback

        # parse through options
        self.stix_validator = options.get('stix_validator', False)
        self.cybox_default = options.get('cybox_default', True)

        self.properties = observable.properties

    @staticmethod
    def _get_value(obj, ds_key, transformer):
        """
        Get value from source object, transforming if specified

        :param obj: the input object we are translating to STIX
        :param ds_key: the property from the input object
        :param transformer: the transform to apply to the property value (can be None)
        :return: the resulting STIX value
        """
        if ds_key not in obj:
            print('{} not found in object'.format(ds_key))
            return None
        ret_val = obj[ds_key]
        # Is this getting his with a none-type value?
        if ret_val and transformer is not None:
            return transformer.transform(ret_val)
        return ret_val

    @staticmethod
    def _add_property(obj, key, stix_value, group=False):
        """
        Add stix_value to dictionary based on the input key, the key can be '.'-separated path to inner object

        :param obj: the dictionary we are adding our key to
        :param key: the key to add
        :param stix_value: the STIX value translated from the input object
        """

        split_key = key.split('.')
        child_obj = obj
        parent_props = split_key[0:-1]
        for prop in parent_props:
            if prop not in child_obj:
                child_obj[prop] = {}
            child_obj = child_obj[prop]

        if split_key[-1] not in child_obj.keys():
            child_obj[split_key[-1]] = stix_value
        elif group is True:  # Mapping of multiple data fields to single STIX object field. Ex: Network Protocols
            if (isinstance(child_obj[split_key[-1]], list)):
                child_obj[split_key[-1]].extend(stix_value)                      # append to existing list

    @staticmethod
    def _handle_cybox_key_def(key_to_add, observation, stix_value, obj_name_map, obj_name, group=False):
        """
        Handle the translation of the input property to its STIX CybOX property

        :param key_to_add: STIX property key derived from the mapping file
        :param observation: the the STIX observation currently being worked on
        :param stix_value: the STIX value translated from the input object
        :param obj_name_map: the mapping of object name to actual object
        :param obj_name: the object name derived from the mapping file
        """
        obj_type, obj_prop = key_to_add.split('.', 1)
        objs_dir = observation['objects']

        if obj_name in obj_name_map:
            obj = objs_dir[obj_name_map[obj_name]]
        else:
            obj = {'type': obj_type}
            obj_dir_key = str(len(objs_dir))
            objs_dir[obj_dir_key] = obj
            if obj_name is not None:
                obj_name_map[obj_name] = obj_dir_key
        DataSourceObjToStixObj._add_property(obj, obj_prop, stix_value, group)

    @staticmethod
    def _valid_stix_value(props_map, key, stix_value):
        """
        Checks that the given STIX value is valid for this STIX property

        :param props_map: the map of STIX properties which contains validation attributes
        :param key: the STIX property name
        :param stix_value: the STIX value translated from the input object
        :return: whether STIX value is valid for this STIX property
        :rtype: bool
        """

        #  Causing a couple of failing tests in MSATP
        if stix_value is None or stix_value == '':
            print("Removing invalid value '{}' for {}".format(stix_value, key))
            return False
        elif key in props_map and 'valid_regex' in props_map[key]:
            pattern = re.compile(props_map[key]['valid_regex'])
            if not pattern.match(str(stix_value)):
                return False
        return True

    def _transform(self, object_map, observation, ds_map, ds_key, obj):

        to_map = obj[ds_key]

        if ds_key not in ds_map:
            print('{} is not found in map, skipping'.format(ds_key))
            return

        if isinstance(to_map, dict):
            print('{} is complex; descending'.format(to_map))
            # If the object is complex we must descend into the map on both sides
            for key in to_map.keys():
                self._transform(object_map, observation, ds_map[ds_key], key, to_map)
            return

        generic_hash_key = ''

        # get the stix keys that are mapped
        ds_key_def_obj = ds_map[ds_key]
        if isinstance(ds_key_def_obj, list):
            ds_key_def_list = ds_key_def_obj
        else:
            # Use callback function to run module-specific logic to handle unknown filehash types
            if self.callback:
                try:
                    generic_hash_key = self.callback(obj, ds_key, ds_key_def_obj['key'], self.options)
                except(Exception):
                    return

            ds_key_def_list = [ds_key_def_obj]

        for ds_key_def in ds_key_def_list:
            if ds_key_def is None or 'key' not in ds_key_def:
                print('{} is not valid (None, or missing key)'.format(ds_key_def))
                continue

            if generic_hash_key:
                key_to_add = generic_hash_key
            else:
                key_to_add = ds_key_def['key']

            transformer = self.transformers[ds_key_def['transformer']] if 'transformer' in ds_key_def else None

            group = False
            if ds_key_def.get('cybox', self.cybox_default):
                object_name = ds_key_def.get('object')
                if 'references' in ds_key_def:
                    references = ds_key_def['references']
                    if isinstance(references, list):
                        stix_value = []
                        for ref in references:
                            val = object_map.get(ref)
                            if not DataSourceObjToStixObj._valid_stix_value(self.properties, key_to_add, val):
                                continue
                            stix_value.append(val)
                        if not stix_value:
                            continue
                    else:
                        stix_value = object_map.get(references)
                        if not DataSourceObjToStixObj._valid_stix_value(self.properties, key_to_add, stix_value):
                            continue
                else:
                    stix_value = DataSourceObjToStixObj._get_value(obj, ds_key, transformer)
                    if not DataSourceObjToStixObj._valid_stix_value(self.properties, key_to_add, stix_value):
                        continue

                # Group Values
                if 'group' in ds_key_def:
                    group = True
                DataSourceObjToStixObj._handle_cybox_key_def(key_to_add, observation, stix_value, object_map, object_name, group)
            else:
                # get the object name defined for custom attributes
                if 'object' in ds_key_def:
                    object_name = ds_key_def.get('object')
                    # get the value mapped
                    if 'value' in ds_key_def:
                        stix_value = ds_key_def['value']
                    # get the value from mapped key
                    elif 'ds_key' in ds_key_def:
                        ds_key = ds_key_def['ds_key']
                        stix_value = DataSourceObjToStixObj._get_value(obj, ds_key, transformer)
                    if not DataSourceObjToStixObj._valid_stix_value(self.properties, key_to_add, stix_value):
                        continue
                    DataSourceObjToStixObj._handle_cybox_key_def(key_to_add, observation, stix_value, object_map, object_name, group)
                else:
                    stix_value = DataSourceObjToStixObj._get_value(obj, ds_key, transformer)
                    if not DataSourceObjToStixObj._valid_stix_value(self.properties, key_to_add, stix_value):
                        continue

                    DataSourceObjToStixObj._add_property(observation, key_to_add, stix_value, group)

    def transform(self, obj):
        """
        Transforms the given object in to a STIX observation based on the mapping file and transform functions

        :param obj: the datasource object that is being converted to stix
        :return: the input object converted to stix valid json
        """
        NUMBER_OBSERVED_KEY = 'number_observed'
        object_map = {}
        stix_type = 'observed-data'
        ds_map = self.ds_to_stix_map

        observation = {
            'id': stix_type + '--' + str(uuid.uuid4()),
            'type': stix_type,
            'created_by_ref': self.identity_id,
            'created': "{}Z".format(datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%S.%f')[:-3]),
            'modified': "{}Z".format(datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%S.%f')[:-3]),
            'objects': {}
        }

        # create normal type objects
        if isinstance(obj, dict):
            for ds_key in obj.keys():
                self._transform(object_map, observation, ds_map, ds_key, obj)
        else:
            print("Not a dict: {}".format(obj))
        
        # Add required property to the observation if it wasn't added via the mapping
        if NUMBER_OBSERVED_KEY not in observation:
            observation[NUMBER_OBSERVED_KEY] = 1

        # Validate each STIX object
        if self.stix_validator:
            validated_result = validate_instance(observation)
            print_results(validated_result)

        return observation



from abc import ABCMeta, abstractmethod
from enum import Enum


class Status(Enum):
    COMPLETED = 'COMPLETED'
    ERROR = 'ERROR'
    CANCELED = 'CANCELED'
    TIMEOUT = 'TIMEOUT'
    RUNNING = 'RUNNING'


class BaseStatusConnector(object, metaclass=ABCMeta):
    @abstractmethod
    def create_status_connection(self, search_id):
        """
        Creates a connection to the specified datasource to determine the status of a given query

        Args:
            search_id (str): The datasource query ID.

        Returns:
            dict: The return value.
                keys:
                    success (bool): True or False
                    status (enum 'Status'): 
                    progress (int): percentage of progress (0-100)
                    error (str): error message (when success=False)
        """
        pass



from stix_shifter_utils.utils.error_mapper_base import ErrorMapperBase
from stix_shifter_utils.utils.error_response import ErrorCode
from stix_shifter_utils.stix_translation.src.utils.exceptions import DataMappingException, StixValidationException, UnsupportedDataSourceException, TranslationResultException
from stix_shifter_utils.stix_translation.src.patterns.errors import SearchFeatureNotSupportedError

error_mapping = {
    NotImplementedError.__name__: [ErrorCode.TRANSLATION_NOTIMPLEMENTED_MODE, 'wrong parameter'],
    DataMappingException.__name__: [ErrorCode.TRANSLATION_MAPPING_ERROR, 'data mapping error'],
    StixValidationException.__name__: [ErrorCode.TRANSLATION_STIX_VALIDATION, 'stix validation error'],
    SearchFeatureNotSupportedError.__name__: [ErrorCode.TRANSLATION_NOTSUPPORTED, 'search feature is not supported'],
    TranslationResultException.__name__: [ErrorCode.TRANSLATION_RESULT, 'result translation error'],
    UnsupportedDataSourceException.__name__: [ErrorCode.TRANSLATION_NOTIMPLEMENTED_MODE, 'unsupported datasource']
}


class ErrorMapper():

    DEFAULT_ERROR = ErrorCode.TRANSLATION_MODULE_DEFAULT_ERROR

    @staticmethod
    def set_error_code(data_dict, return_obj):
        exception = None
        if 'exception' in data_dict:
            exception = data_dict['exception']

        error_code = ErrorMapper.DEFAULT_ERROR
        error_message = 'Error when converting STIX pattern to data source query'

        if exception is not None:
            exception_type = type(exception).__name__
            print("received exception => {}: {}".format(exception_type, exception))
            if exception_type in error_mapping:
                error_code = error_mapping[exception_type][0]
                error_message = error_mapping[exception_type][1]
                exception_message = str(exception)
                if (len(exception_message) > 0):
                    if len(error_message) > 0:
                        error_message += ' : '
                    error_message += exception_message

        ErrorMapperBase.set_error_code(return_obj, error_code, message=error_message)



from abc import ABCMeta, abstractmethod


class BaseResultsConnector(object, metaclass=ABCMeta):
    @abstractmethod
    def create_results_connection(self, search_id, offset, length):
        """
        Creates a connection to the specified datasource to retrieve query results

        Args:
            search_id (str): The datasource query ID.
            offset: data offset to start fetch from.
            length: data length to fetch

        Returns:
            dict: The return value.
                keys:
                    success (bool): True or False
                    data (str): The query result data
                    error (str): error message (when success=False)
        """
        pass



from .base_ping_connector import BasePingConnector
from .base_query_connector import BaseQueryConnector
from .base_status_connector import BaseStatusConnector
from .base_delete_connector import BaseDeleteConnector
from .base_results_connector import BaseResultsConnector


class BaseConnector:
    def create_query_connection(self, query):
        """
        Creates a connection to the specified datasource to send a query

        Args:
            query (str): The datasource query.

        Returns:
            dict: The return value.
                keys:
                    success (bool): True or False
                    search_id (str): query ID
                    error (str): error message (when success=False)
        """
        raise NotImplementedError()

    def create_status_connection(self, search_id):
        """
        Creates a connection to the specified datasource to determine the status of a given query

        Args:
            search_id (str): The datasource query ID.

        Returns:
            dict: The return value.
                keys:
                    success (bool): True or False
                    status (enum 'Status'): 
                    progress (int): percentage of progress (0-100)
                    error (str): error message (when success=False)
        """
        raise NotImplementedError()

    def create_results_connection(self, search_id, offset, length):
        """
        Creates a connection to the specified datasource to retrieve query results

        Args:
            search_id (str): The datasource query ID.
            offset: data offset to start fetch from.
            length: data length to fetch

        Returns:
            dict: The return value.
                keys:
                    success (bool): True or False
                    data (str): The query result data
                    error (str): error message (when success=False)
        """
        raise NotImplementedError()

    def delete_query_connection(self, search_id):
        """
        Deletes a query from the specified datasource

        Args:
            search_id (str): The datasource query ID.

        Returns:
            dict: The return value.
                keys:
                    success (bool): True or False
                    error (str): error message (when success=False)
        """
        raise NotImplementedError()

    def ping_connection(self):
        """
        Sends a basic request to the datasource to confirm we are connected and authenticated

        Args:
            None.

        Returns:
            dict: The return value.
                keys:
                    success (bool): True or False
                    error (str): error message (when success=False)
        """
        raise NotImplementedError()



from setuptools import find_packages
# To use a consistent encoding
from codecs import open
import sys
import shutil
import subprocess
import json
import io
import os

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

def fill_connectors(projects, modules_path):
    modules = [ name for name in os.listdir(modules_path) if (os.path.isdir(os.path.join(modules_path, name) ) and (not name.startswith('__')) )]
    for module in modules:
        if not os.path.isfile(os.path.join(modules_path, module, 'SKIP.ME')):
            projects['stix_shifter_modules_'+module] = ['stix_shifter_modules/'+module]

mode = '1'
if 'MODE' in os.environ:
    mode = os.environ['MODE']

#DO NOT remove spaces around the equal sign at the line below
version = '1.0.0'
if 'VERSION' in os.environ:
    version = os.environ['VERSION']

if mode=='1':
    projects = {
        "stix_shifter": [
            'stix_shifter_utils',
            'stix_shifter', 
            'stix_shifter_modules'
            ]
    }
elif mode=='3':
    projects = {
        "stix_shifter_utils": ["stix_shifter_utils"],
        "stix_shifter": ["stix_shifter"],
        "stix_shifter_modules": ["stix_shifter_modules"],
    }
elif mode=='N':
    projects = {
        "stix_shifter_utils": ["stix_shifter_utils"],
        "stix_shifter": ["stix_shifter"],
    }
    fill_connectors(projects, "stix_shifter_modules")
else:
    print('Unexpected value in MODE environment variable: %s' % mode)
    print('Allowed values: 1|3|N')
    exit(1)

for project_name in projects.keys():
    src_folders = projects[project_name]
    print('processing project_name %s' % project_name)

    #Prepare packages 
    packages_include = []
    for src_folder in src_folders:
        packages_include.append(src_folder.replace('/','.'))
        packages_include.append(src_folder.replace('/','.') + '.*')
    print('packages_include: %s' % packages_include)
    packages = find_packages(include=packages_include)
    print('packages: %s' % packages)

    #Prepare requires list
    install_requires = set()
    requirements_files = []
    for src_folder in src_folders:
        for r, d, f in os.walk(src_folder):
            for file in f:
                if 'requirements.txt'==file and not os.path.isfile(os.path.join(r, 'SKIP.ME')):
                    requirements_files.append(os.path.join(r, file))
    print('requirements_files: %s' % requirements_files)
    for requirements_file in requirements_files:
        with open(requirements_file) as f:
            lines = f.readlines()
        lines = [x.strip() for x in lines]
        lines = list(filter(lambda s: (not s.startswith('#')) and len(s)>0, lines))
        install_requires.update(lines)
    install_requires = list(install_requires)
    print('install_requires: %s' % install_requires)

    #Prepare entry points
    entry_points = {}
    entry_points_items = []
    for src_folder in src_folders:
        entry_point_path = os.path.join(src_folder, 'scripts', src_folder+'.py')
        if os.path.exists(entry_point_path):
            entry_points_items.append('%s=%s.scripts.%s:main' % (project_name.replace('_','-'),project_name,project_name))
    if len(entry_points_items) > 0:
        entry_points = {  # Optional
            'console_scripts': entry_points_items
            }
    print('entry_points: %s' % entry_points)

    #Prepare setup params
    params = {
        'name': project_name,  # Required
        'version': version,  # Required
        'description': 'Tools and interface to translate STIX formatted results and queries to different data source formats and to set up appropriate connection strings for invoking and triggering actions in openwhisk',  # Required
        'long_description': long_description,  # Optional
        'long_description_content_type': 'text/markdown',  # Optional (see note above)
        'url': 'https://github.com/opencybersecurityalliance/stix-shifter',  # Optional
        'author': 'ibm',
        'author_email': '',  # Optional
        # https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'classifiers': [  # Optional
            'License :: OSI Approved :: Apache Software License',
            'Programming Language :: Python :: 3.6',
        ],
        'keywords': 'datasource stix translate transform transmit',  # Optional
        'packages': packages, # Required
        'install_requires': install_requires,
        'include_package_data': True,
        'entry_points': entry_points,  # Optional
        'project_urls': {  # Optional
            'Source': 'https://github.com/opencybersecurityalliance/stix-shifter',
        },
    }

    #Prepare MANIFEST.in, include json files
    shutil.rmtree('MANIFEST.in', ignore_errors=True)
    shutil.copyfile('build_templates/MANIFEST.in', 'MANIFEST.in')
    json_include_lines = []
    for src_folder in src_folders:
        json_search_path = src_folder
        for r, d, f in os.walk(json_search_path):
            for file in f:
                if '.json' in file:
                    json_include_lines.append('include '+os.path.join(r, file)+' \n')
    with open('MANIFEST.in', 'a') as out_file:
        out_file.writelines(json_include_lines)
    
    #Run setup()
    proc_params = ['python3','build_templates/setup_one.py']
    proc_params.extend(sys.argv[1:])
    proc_params.append(json.dumps(params))
    proc = subprocess.Popen(proc_params,stdout=subprocess.PIPE)
    for line in io.TextIOWrapper(proc.stdout, encoding="utf-8"):
        if not line:
            break
        print(line.rstrip())

    #Cleanup
    cleanup_file_list = ['build', 'MANIFEST.in', project_name + '.egg-info']
    for cleanup_file in cleanup_file_list:
        if os.path.exists(cleanup_file):
            if os.path.isdir(cleanup_file):
                shutil.rmtree(cleanup_file)
            else:
                os.remove(cleanup_file)

    print('---------------------------------')


import os

def process_dialects(cli_module, options):
    module, cli_dialects = __split_module_dialects(cli_module)
    if 'dialects' not in options:
        options['dialects'] = []
    options_dialects = options['dialects']
    
    dialects = list(set(cli_dialects+options_dialects))
    return module, dialects

def __split_module_dialects(module_dialects):
    module = None
    dialects = []
    if ':' in module_dialects:
        dialects = module_dialects.split(':')
        module = dialects.pop(0)
    else:
        module = module_dialects
        dialects = []
    return module, dialects

def dialect_list(module):
    dialects_path = os.path.join(os.path.dirname(__file__), f'../../stix_shifter_modules/{module}/stix_translation/json')
    ENDING = '_from_stix_map.json'
    dialects = []
    if os.path.isdir(dialects_path):
        files = [ name for name in os.listdir(dialects_path) if (os.path.isfile(os.path.join(dialects_path, name)) and (name.endswith(ENDING)) )]
        dialects = list(map(lambda f: f[:-len(ENDING)], files))
    if not dialects:
        dialects = ['default']
    return dialects


import logging

import dateutil.parser
from antlr4 import CommonTokenStream, ParseTreeWalker, InputStream
from .grammar import STIXPatternListener, STIXPatternParser, STIXPatternLexer
from .pattern_objects import ObservationExpression, CombinedComparisonExpression, ObservationOperators, \
    ComparisonExpressionOperators, ComparisonComparators, SetValue, ComparisonExpression, CombinedObservationExpression, Pattern, Qualifier, StartStopQualifier

logger = logging.getLogger(__name__)


class STIXQueryBuilder(STIXPatternListener):

    query_comparators = {
        'and': ComparisonExpressionOperators.And,
        # 'not': ComparisonExpressionOperators.Not,
        'or': ComparisonExpressionOperators.Or,
    }

    def __init__(self) -> None:
        self._stack = []
        # self._saved_stack is used to save the main stack, when a second stack is used to parse a sequence
        # of indterminate length (e.g. a STIX Patterning Set).
        self._saved_stack = self._stack

    def push(self, x) -> None:
        self._stack.append(x)

    def pop(self):
        return self._stack.pop()

    def exitObjectPath(self, ctx) -> None:
        logger.debug("{} {} {}".format("ObjectPath", ctx, ctx.getText()))
        # Single quotes are valid for any path but required for those with dashes.
        # They're not really relevant once they're parsed out, so just remove them
        self.push(ctx.getText().replace("'", ""))

    def exitPropTestLike(self, ctx: STIXPatternParser.PropTestLikeContext) -> None:
        logger.debug("{} {} {}".format("PropTestLike", ctx, ctx.getText()))
        value = self.pop()
        object_path = self.pop()
        negated = ctx.NOT()
        self.push(ComparisonExpression(object_path, value, ComparisonComparators.Like, negated=negated))

    def exitPrimitiveLiteral(self, ctx) -> None:
        logger.debug("{} {} {}".format("PrimitiveLiteral", ctx, ctx.getText()))
        # This can be TimestampLiteral or BoolLiteral.  These should be handed with more specific targets, e.g.
        #   exitTimestampLiteral and exitBoolLiteral.
        # self.push(ctx.getText().strip("'"))

    def exitPropTestRegex(self, ctx: STIXPatternParser.PropTestRegexContext) -> None:
        logger.debug("{} {} {}".format("PropTestRegex", ctx, ctx.getText()))
        logger.debug("stack={}".format(self._stack))
        value = self.pop()
        object_path = self.pop()
        negated = ctx.NOT()
        self.push(ComparisonExpression(object_path, value, ComparisonComparators.Matches, negated=negated))

    def exitPropTestEqual(self, ctx: STIXPatternParser.PropTestEqualContext) -> None:
        logger.debug("{} {} {} | stack={}".format("PropTestEqual", ctx, ctx.getText(), self._stack))
        value = self.pop()
        object_path = self.pop()
        negated = ctx.NOT()

        if ctx.EQ():
            comparator = ComparisonComparators.Equal
        elif ctx.NEQ():
            comparator = ComparisonComparators.NotEqual
        else:
            raise RuntimeWarning("Unrecognized Equality Comparator")
        self.push(ComparisonExpression(object_path, value, comparator, negated=negated))

    def exitPropTestOrder(self, ctx: STIXPatternParser.PropTestOrderContext) -> None:
        logger.debug("{} {} {} | stack={}".format("PropTestOrder", ctx, ctx.getText(), self._stack))
        value = self.pop()
        object_path = self.pop()
        negated = str(ctx.NOT()) == "NOT"

        if ctx.GT():
            comparator = ComparisonComparators.GreaterThan
        elif ctx.LT():
            comparator = ComparisonComparators.LessThan
        elif ctx.GE():
            comparator = ComparisonComparators.GreaterThanOrEqual
        elif ctx.LE():
            comparator = ComparisonComparators.LessThanOrEqual
        else:
            raise RuntimeWarning("Unrecognized Ordering Comparator")

        self.push(ComparisonExpression(object_path, value, comparator=comparator, negated=negated))

    def exitPropTestSet(self, ctx: STIXPatternParser.PropTestSetContext) -> None:
        logger.debug("{} {} {} | stack={}".format("PropTestSet", ctx, ctx.getText(), self._stack))
        vals = self.pop()
        object_path = self.pop()
        negated = str(ctx.NOT()) == "NOT"

        self.push(ComparisonExpression(object_path, vals, ComparisonComparators.In, negated=negated))

    def enterSetLiteral(self, ctx: STIXPatternParser.SetLiteralContext) -> None:
        """ Called when Set Literals begin being parsed.  Since the length is unpredictable, replace the stack
        with a SetValue object. SetValue is also a stack (implements .pop() and .push()).  When the SetLiteral is
        done being parsed, the original stack will be restored, and the SetValue pushed onto it."""
        self._saved_stack = self._stack
        self._stack = SetValue()

    def exitSetLiteral(self, ctx: STIXPatternParser.SetLiteralContext):
        """Called when SetLiterals are done being parsed. Restore the stack and push the built SetValue object onto
        the top of the stack so it can be used in a comparison expression."""
        if not isinstance(self._stack, SetValue):
            raise RuntimeError("Exiting Set Literal but current stack isn't Set Object")
        else:
            set_value = self._stack
            set_value.close()
            self._stack = self._saved_stack
            self.push(set_value)

    def exitOrderableLiteral(self, ctx: STIXPatternParser.OrderableLiteralContext):
        """  Can be IntPosLiteral, IntNegLiteral, FloatPosLiteral, FloatNegLiteral, stringLiteral, BinaryLiteral, HexLiteral, TimestampLiteral """
        logger.debug("{} {} {}".format("OrderableLiteral", ctx, ctx.getText()))
        if ctx.stringLiteral():
            pass  # Strings should have already been pushed onto the stack
        elif ctx.IntPosLiteral() or ctx.IntNegLiteral():
            self.push(int(ctx.getText()))
        elif ctx.FloatPosLiteral() or ctx.FloatNegLiteral():
            self.push(float(ctx.getText()))
        elif ctx.BinaryLiteral():
            # Leave these as strings for now
            self.push(ctx.getText())
        elif ctx.HexLiteral():
            # Leave these as strings for now
            self.push(ctx.getText())
        elif ctx.TimestampLiteral():  # TODO: is this the right way?
            ts = dateutil.parser.parse(ctx.getText().strip("'").strip("t"))
            self.push(ts)

    def exitComparisonExpressionAnded(self, ctx):
        logger.debug("{} {} {}".format("ComparisonExpressionAnded", ctx, ctx.getText()))
        exp1 = self.pop()
        exp2 = self.pop()
        self.push(CombinedComparisonExpression(exp1, exp2, ComparisonExpressionOperators.And))

    def exitComparisonExpressionOred(self, ctx) -> None:
        logger.debug("{} {} {}".format("ComparisonExpressionOred", ctx, ctx.getText()))
        exp1 = self.pop()
        exp2 = self.pop()
        self.push(CombinedComparisonExpression(exp1, exp2, ComparisonExpressionOperators.Or))

    def exitStringLiteral(self, ctx: STIXPatternParser.StringLiteralContext) -> None:
        logger.debug("{} {} {}".format("String", ctx, ctx.getText()))
        self.push(ctx.getText().strip("'").replace("\\\\", "\\"))

    def exitObservationExpressionSimple(self, ctx):
        # Roughly analogous to CQL DataModelQuery
        logger.debug("{} {} {}".format("ObservationExpressionSimple", ctx, ctx.getText()))
        comparison_expression = self.pop()

        # Done with this observation expression, the next one might have a different object type
        # object_name = self._current_object_type
        # self._current_object_type = None

        logger.debug("Current Parser Stack: {}".format(self._stack))
        # logger.debug("Building DMQ with object_name={}, action={}, query={}".format(object_name, self.action, query))
        # self.push(DataModelQuery(object_name=object_name, action=self.action, query=query))
        self.push(ObservationExpression(comparison_expression))

    def exitObservationExpressionAnd(self, ctx: STIXPatternParser.ObservationExpressionAndContext):
        logger.debug("{} {} {}".format("ObservationExpressionAnd", ctx, ctx.getText()))
        if ctx.AND():
            expr2 = self.pop()
            expr1 = self.pop()
            operator = ObservationOperators.And
            self.push(CombinedObservationExpression(expr1, expr2, operator))

    def exitObservationExpressionOr(self, ctx: STIXPatternParser.ObservationExpressionOrContext):
        logger.debug("{} {} {}".format("ObservationExpressionOr", ctx, ctx.getText()))
        if ctx.OR():  # Check if there's OR'd Observation Expressions
            expr2 = self.pop()
            expr1 = self.pop()
            operator = ObservationOperators.Or
            self.push(CombinedObservationExpression(expr1, expr2, operator))

    def exitObservationExpressionStartStop(self, ctx: STIXPatternParser.ObservationExpressionStartStopContext):
        logger.debug("{} {} {}".format("ObservationExpressionStartStop", ctx, ctx.getText()))
        qualifier = ctx.startStopQualifier()
        qualifier_text = qualifier.getText()  # Ex: "START'2016-06-01T00:00:00Z'STOP'2016-06-01T01:11:11Z'"
        observation = ctx.observationExpression()
        expression = self.pop()
        start_time = str(qualifier.TimestampLiteral(i=0))
        stop_time = str(qualifier.TimestampLiteral(i=1))
        observation_expression_with_qualifier = StartStopQualifier(qualifier_text, expression, start_time, stop_time)
        self.push(observation_expression_with_qualifier)

    def exitObservationExpressions(self, ctx: STIXPatternParser.ObservationExpressionsContext):
        logger.debug("{} {} {}".format("ObservationExpressions", ctx, ctx.getText()))
        if ctx.FOLLOWEDBY():
            expr2 = self.pop()
            expr1 = self.pop()
            operator = ObservationOperators.FollowedBy
            self.push(CombinedObservationExpression(expr1, expr2, operator))

    def exitPattern(self, ctx):
        logger.debug("{} {} {}".format("Pattern", ctx, ctx.getText()))
        observation_expression = self.pop()
        self.push(Pattern(observation_expression))

    def exitPropTestIsSuperset(self, ctx: STIXPatternParser.PropTestIsSupersetContext) -> None:
        logger.debug("{} {} {}".format("exitPropTestIsSuperset", ctx, ctx.getText()))
        value = self.pop()
        object_path = self.pop()
        self.push(ComparisonExpression(object_path, value, ComparisonComparators.IsSuperSet))

    def exitPropTestIsSubset(self, ctx: STIXPatternParser.PropTestIsSubsetContext) -> None:
        logger.debug("{} {} {}".format("propTestIsSubset", ctx, ctx.getText()))
        value = self.pop()
        object_path = self.pop()
        negated = ctx.NOT()
        self.push(ComparisonExpression(object_path, value, ComparisonComparators.IsSubSet, negated=negated))

# copied from CASCADE data_model (defined twice)


class InvalidFieldError(KeyError):
    pass


class InvalidActionError(ValueError):
    pass


class InvalidObjectError(ValueError):
    pass


class ParserError(ValueError):
    pass


def generate_query(query_string):
    try:
        lexer = STIXPatternLexer(InputStream(query_string))
        stream = CommonTokenStream(lexer)
        parser = STIXPatternParser(stream)
        builder = STIXQueryBuilder()
        tree = parser.pattern()
        walker = ParseTreeWalker()
        walker.walk(builder, tree)
        query = builder.pop()
        logger.debug(query)
    except Exception as e:
        raise ParserError(e)

    return query



from stix_shifter_utils.stix_translation.src.utils.exceptions import DataMappingException
from stix_shifter_utils.modules.base.stix_translation.base_data_mapper import BaseDataMapper
from os import path


class CarDataMapper(BaseDataMapper):

    def __init__(self, options={}, dialect=None):
        super().__init__(dialect)
        self.options = options
        mapping_json = options.get('mapping')
        basepath = path.dirname(__file__)
        print('mapping_json: ' + str(mapping_json))
        self.map_data = mapping_json or self.fetch_mapping(basepath)

    def map_object(self, stix_object_name):
        if self.map_data.get(stix_object_name):
            return self.map_data[stix_object_name]["car_type"]
        else:
            raise DataMappingException("Unable to map object `{}` into CAR".format(stix_object_name))


mapper_class = CarDataMapper



from stix_shifter_utils.stix_translation.src.utils.exceptions import DataMappingException
from stix_shifter_utils.modules.base.stix_translation.base_data_mapper import BaseDataMapper
from os import path


class CimDataMapper(BaseDataMapper):

    FIELDS = {
        "default": [
            "src_ip",
            "src_port",
            "src_mac",
            "src_ipv6",
            "dest_ip",
            "dest_port",
            "dest_mac",
            "dest_ipv6",
            "file_hash",
            "user",
            "url",
            "protocol"
        ]
    }

    def __init__(self, options={}, dialect=None):
        super().__init__(dialect)
        self.options = options
        if 'select_fields' in options:
            self.FIELDS = options['select_fields']
        mapping_json = options.get('mapping')
        basepath = path.dirname(__file__)
        self.map_data = mapping_json or self.fetch_mapping(basepath)

    # TODO:
    # This mapping is not super straightforward. It could use the following improvements:
    # * Registry keys need to pull the path apart from the key name, I believe. Need to investigate with Splunk
    # * Need to validate that the src and dest aliases are working
    # * Need to add in the static attributes, like the `object_category` field
    # * Need to verify "software" == "inventory"
    # * Need to figure out network traffic when it's for web URLs
    # * Hashes are kind of hacky, just hardcoded. Probably needs to be a regex

    def map_object(self, stix_object_name):
        if stix_object_name in self.map_data and self.map_data[stix_object_name] != None:
            return self.map_data[stix_object_name]["cim_type"]
        else:
            raise DataMappingException("Unable to map object `{}` into CIM".format(stix_object_name))

